"""
Copied from the original training simulation code and adapted.

# The MIT License (MIT)
# Copyright (c) 2021 Ashwin De Silva
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial
# portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
# TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
"""
import numpy as np


def simulate_gauss(size, A, mu_x, mu_y, sigma_x, sigma_y):
    """
    creates a phase map by mixing gaussian blobs
    """
    x = np.arange(0, size[1], 1) # cols
    y = np.arange(0, size[0], 1) # rows

    xx, yy = np.meshgrid(x, y, sparse=True)

    I = np.zeros(size)

    ## mix randomly shaped and placed gaussian blobs
    for i in range(len(sigma_x)):
      a = (xx-mu_x[i])**2/(2*sigma_x[i]**2) + (yy-mu_y[i])**2/(2*sigma_y[i]**2)
      I += A[i]*np.exp(-a)

    return I

def simulate_ramp(size, m_x, m_y, C): 
    """
    Creates a phase ramp
    """
    x = np.arange(0, size[1], 1) # cols
    y = np.arange(0, size[0], 1) # rows

    # meshgrid
    xx, yy = np.meshgrid(x, y, sparse=True)

    ## add ramp phase with random gradients and shifts
    I = m_x*xx + m_y*yy + C
  
    return I

def wrap(phi):
  """
  wraps the true phase signal within [-pi, pi]
  """
  return np.angle(np.exp(1j*phi))

def rescale(im, interval):
  """
  mini-max rescales the input image
  """
  im_std = (im - im.min()) / (im.max() - im.min())
  im_scaled = im_std * (interval[1] - interval[0]) + interval[0]
  return im_scaled

def fracsurf(Nx, Ny, beta, P0 = 1):
    '''
    generates a fractal surface simulation
    # simulation of atmospheric turbulent phase
    # Adapted from fracsurf.m 
    # http://doris.tudelft.nl/software/download.html.081208
    # Bert Kampes, Ramon Hanssen, Zbigniew Perski. 
    # Radar Interferometry with Public Domain Tools.
    # In: Proceedings of FRINGE 2003, December 1-5, Frascati, Italy, 2003.
    '''
    # Simulate a uniform random signal
    h  = np.random.rand(Ny, Nx)
    H  = np.fft.fftshift(np.fft.fft2(h))
    
    # simulate wavenumber
    x = np.arange(-(Nx//2), Nx/2)
    y = np.arange(-(Ny//2), Ny/2)
    X,Y = np.meshgrid(x, y)
    k = np.sqrt(X*X + Y*Y)
    
    # add 1 to beta because we use a 1D slice of 2D spectrum
    beta = beta + 1
    
    # sqrt of spectrum
    noemer = k**(beta/2)
    noemer[noemer == 0] = 1 
    Hnew = np.sqrt(P0) * H / noemer
    
    # inverse fft
    fsurf = np.abs(np.fft.ifft2(np.fft.ifftshift(Hnew)))
    
    # demean
    fsurf -= np.mean(fsurf)
    
    return fsurf


def simulate_images(num_gaussians=3, safety_margin=0.12,
                   invert_gaussians=False, gauss_scale=3*2*np.pi,
                   ramp_scale=3*2*np.pi,
                   noise_std=0, atmo_scale=0, size=(256,256)):
    """
    Simulate gaussian mixture, ramp, atmospheric phase, noise.
    
    Args:
    -----
    num_gaussians: integer, number of gaussians in the simulation, 
                    if in [2, 5], will be inside the range used for training
    safety_margin: float, fraction of pixels as safety margins from the border of the 
                   image, to restrict the area where the center of the gaussians are taken
                   if = 0.08, this corresponds to the value set by the author
    invert_guaussians: bool, if True, half of the gaussians are inverted (amplitude multiplied by -1)
    gauss_scale: float, gaussian will be scaled to [-gauss_scale/2, gauss_scale/2]
    ramp_scale: float, the phase ramp will be rescaled to [-ramp_scale/2, ramp_scale/2]
    noise_std: float, noise standard deviation, should be less than pi. (pi would induce a totally random phase)
    atmo_scale: float, atmospheric phase will be rescaled to [-atmo_scale/2, atmo_scale/2]
    size: tuple (h, w), output image size
    
    Returns:
    --------
    I_gauss: (h, w) array Gaussian mixture image simulation
    I_ramp: (h, w) array Ramp image simulation
    I_atmo: (h, w) array Atmospheric image simulation
    I_noise: (h, w) array Noise image simulation
    
    """

    h, w = size

    I_gauss = np.zeros(size)

    if num_gaussians and gauss_scale:
        # start by creating gaussians
        A = np.random.uniform(0.2, 1, size=num_gaussians)
        
        if invert_gaussians:
            # invert the sign of half the gaussians
            A[::2] *= -1
                
        mu_x = np.random.randint(safety_margin, w - 1 - int(safety_margin*w), num_gaussians)
        mu_y = np.random.randint(safety_margin, h - 1 - int(safety_margin*h), num_gaussians)
        
        sigma_x = np.random.randint(w//25, w//5, num_gaussians)
        sigma_y = np.random.randint(h//25, h//5, num_gaussians)
    
        I_gauss = rescale(simulate_gauss(size, A, mu_x, mu_y, sigma_x, sigma_y),
                          (-gauss_scale/2, gauss_scale/2))

    I_ramp = np.zeros(size)

    if ramp_scale:

        m_x = np.random.uniform(0, 1 )
        m_y = np.random.uniform(0, 1)
        C = np.random.uniform(0, 1)
        I_ramp = rescale(simulate_ramp(size, m_x, m_y, C),
                         (-ramp_scale/2, ramp_scale/2))

    I_atmo = np.zeros(size)
    if atmo_scale:
        fsurf = fracsurf(w, h, 5/3, 1)
        I_atmo = rescale(fsurf, (-atmo_scale/2, atmo_scale/2))

    I_noise = np.zeros(size)

    if noise_std:
        I_noise = np.random.normal(0, noise_std, size=size)

    return I_gauss.astype("float32"), I_ramp.astype("float32"), I_atmo.astype("float32"), I_noise.astype("float32")